import os, sys
import numpy as np
import GWDALI.lib.Waveforms as wf
import GWDALI.lib.Angles_lib as geo
import GWDALI.lib.Dictionaries as gwdict
import GWDALI.lib.Derivatives_Tensors as gwfunc
import GWDALI.lib.Symmetric_Indexies as sym

from itertools import permutations
from tqdm import trange

deg2rad = np.pi/180

Waveforms, PSD, labels_tex = gwdict.Load_Dictionaries()

def check_permutations(prmt0):
	prmt = []
	for i in range(len(prmt0)):
		p0 = list(prmt0[i])
		if(not p0 in prmt):
			prmt.append(p0)
	return prmt # list type

def Get_Tensors(FreeParams, gw_prms, detectors, dali_method, approximant, hide_info):
	
	#if(hide_info): sys.stdout = open(os.devnull, 'w')

	Np = len(FreeParams)
	rho2 = 0

	Fisher   = np.zeros([Np, Np])
	Doublet3 = np.zeros(Np**3)
	Doublet4 = np.zeros(Np**4)
	Triplet4 = np.zeros(Np**4)
	Triplet5 = np.zeros(Np**5)
	Triplet6 = np.zeros(Np**6)

	num_det = 1
	GwData = [] ; SNR = []
	for det in detectors:

		#-----------------------------------
		# Computing Signal to Noise 
		#-----------------------------------
		# from Derivative_Tensors.py
		h 	  = gwfunc.Signal(gw_prms, det, approximant)
		SNR2  = gwfunc.ScalarProduct(det['freq'], det['Sn'],h,h)
		rho2 += SNR2
		GwData.append(h) ; SNR.append(np.sqrt(SNR2))
		#------------------------------------------------------
		# Computing Fisher Matrix
		#------------------------------------------------------

		if(not hide_info): print("\n\n\t Computing Fisher %s(%d) ..." % (det['name'],num_det))
		for i in range(Np):
			for j in range(i+1):
				if(not hide_info): print('Fisher: [%d-%d]' % (i+1,j+1), end='\r')
				xj = FreeParams[j]
				xi = FreeParams[i]

				Fij = gwfunc.Fisher_ij(xi, xj, gw_prms, det, approximant)
				Fisher[i][j] += Fij
				if(i!=j): Fisher[j][i] += Fij

		brute_force = False
		if(brute_force):
			#------------------------------------------------------
			# Computing Doublet Coefficients (Brute Force)
			#------------------------------------------------------

			#*****************************#*****************************#*****************************
			#*****************************#*****************************#*****************************
			
			if(dali_method in ["Doublet","Triplet"]):
				if(not hide_info): print("\n\n\tComputing Doublet/Triplet Tensors %s(%d)..." % (det['name'],num_det))
				cont = 0
				for i in range(Np):
					xi = FreeParams[i]
					
					for j in range(Np):
						xj = FreeParams[j]
					
						for k in range(Np):
							xk = FreeParams[k]
							
							idx3 = i*Np**2 + j*Np + k
							Doublet3[idx3] += gwfunc.func_doublet3(xi,xj,xk, gw_prms, det, approximant) # [1,2]

							for l in range(Np):
								xl = FreeParams[l]
								if((not hide_info) and dali_method == 'Doublet'):
									print('Doublet: [%d-%d-%d-%d]' % (i+1,j+1,k+1,l+1), end='\r')

								idx4 = i*Np**3 + j*Np**2 + k*Np + l
								Doublet4[idx4] += gwfunc.func_doublet4(xi,xj,xk,xl, gw_prms, det, approximant) # [2,2]

								#*****************************#*****************************#*****************************
								#*****************************#*****************************#*****************************

								if(dali_method == 'Triplet'):
									Triplet4[idx4] += gwfunc.func_triplet4(xi,xj,xk,xl, gw_prms, det, approximant)

									for m in range(Np):
										xm = FreeParams[m]
										idx5 = i*Np**4 + j*Np**3 + k*Np**2 + l*Np + m
										Triplet5[idx5] += gwfunc.func_triplet5(xi,xj,xk,xl,xm, gw_prms, det, approximant)

										for n in range(Np):
											xn = FreeParams[n]
											if(not hide_info):
												print('Triplet: [%d-%d-%d-%d-%d-%d]' % (i+1,j+1,k+1,l+1,m+1,n+1), end='\r')
											
											idx6 = i*Np**5 + j*Np**4 + k*Np**3 + l*Np**2 + m*Np + n
											Triplet6[idx6] += gwfunc.func_triplet6(xi,xj,xk,xl,xm,xn, gw_prms, det, approximant)
								else:
									Triplet4, Triplet5, Triplet6 = None, None, None
			num_det+=1
			#*****************************#*****************************#*****************************
			#*****************************#*****************************#*****************************
			
		else:
			#------------------------------------------------------
			# Checking Symmetries
			#------------------------------------------------------
			idxs_doub, idxs_trip = sym.get_indep_indexies(Np, hide_info)

			if(dali_method in ["Doublet","Triplet"]):
				if(not hide_info): print("\n\n\tComputing Doublet/Triplet Tensors %s..." % det['name'])
				idxsD_12, idxsD_22 = idxs_doub

				for Idx2 in range(len(idxsD_12)):
					i, j, k = idxsD_12[Idx2]
					xi = FreeParams[i]
					xj = FreeParams[j]
					xk = FreeParams[k]
					value = gwfunc.func_doublet3(xi,xj,xk, gw_prms, det, approximant)
					prmt = check_permutations( list(permutations([j,k])) ) ; lenP = len(prmt)
					for p in range(lenP):
						j, k = list(prmt[p])
						IdxP = i*Np**2 + j*Np + k
						if((not hide_info)):
							print('Doublet(1,2): [%d-%d-%d]' % (i+1,j+1,k+1), end='\r')
						Doublet3[IdxP] += value

				if(not hide_info):print('\n')
				for Idx2 in range(len(idxsD_22)):
					i, j, k, l = idxsD_22[Idx2]
					xi = FreeParams[i]
					xj = FreeParams[j]
					xk = FreeParams[k]
					xl = FreeParams[l]
					value = gwfunc.func_doublet4(xi,xj,xk,xl, gw_prms, det, approximant)
					prmt1 = check_permutations( list(permutations([i,j])) ) ; len1 = len(prmt1)
					prmt2 = check_permutations( list(permutations([k,l])) ) ; len2 = len(prmt2)
					for n1 in range(len1):
						for n2 in range(len2):
							i, j, k, l = list(prmt1[n1])+list(prmt2[n2])
							IdxP = i*Np**3 + j*Np**2 + k*Np + l
							if(not hide_info):
								print('Doublet(2,2): [%d-%d-%d-%d]' % (i+1,j+1,k+1,l+1), end='\r')		
							Doublet4[IdxP] += value

				#*****************************#*****************************#*****************************
				if(dali_method == 'Triplet'):
					idxsT_13, idxsT_23, idxsT_33 = idxs_trip

					if(not hide_info):print('\n')
					for Idx3 in range(len(idxsT_13)):
						i, j, k, l = idxsT_13[Idx3]
						xi = FreeParams[i]
						xj = FreeParams[j]
						xk = FreeParams[k]
						xl = FreeParams[l]
						value = gwfunc.func_triplet4(xi,xj,xk,xl, gw_prms, det, approximant)
						prmt = check_permutations( list(permutations([j,k,l])) ) ; lenP = len(prmt)
						for idx in range(lenP):
							j, k, l = list(prmt[idx])
							if(not hide_info):
								print('Triplet(1,3): [%d-%d-%d-%d]' % (i+1,j+1,k+1,l+1), end='\r')
							IdxP = i*Np**3 + j*Np**2 + k*Np + l
							Triplet4[IdxP] += value

					if(not hide_info):print('\n')
					for Idx3 in range(len(idxsT_23)):
						i, j, k, l, m = idxsT_23[Idx3]
						xi = FreeParams[i]
						xj = FreeParams[j]
						xk = FreeParams[k]
						xl = FreeParams[l]
						xm = FreeParams[m]
						value = gwfunc.func_triplet5(xi,xj,xk,xl,xm, gw_prms, det, approximant)
						prmt1 = check_permutations( list(permutations([i,j]))   ) ; len1 = len(prmt1)
						prmt2 = check_permutations( list(permutations([k,l,m])) ) ; len2 = len(prmt2)
						for n1 in range(len1):
							for n2 in range(len2):
								i, j, k, l, m = list(prmt1[n1]) + list(prmt2[n2])
								if(not hide_info):
									print('Triplet(2,3): [%d-%d-%d-%d-%d]' % (i+1,j+1,k+1,l+1,m+1), end='\r')
								IdxP = i*Np**4 + j*Np**3 + k*Np**2 + l*Np + m
								Triplet5[IdxP] += value

					if(not hide_info):print('\n')
					for Idx3 in range(len(idxsT_33)):
						i, j, k, l, m, n = idxsT_33[Idx3]
						xi = FreeParams[i]
						xj = FreeParams[j]
						xk = FreeParams[k]
						xl = FreeParams[l]
						xm = FreeParams[m]
						xn = FreeParams[n]
						value = gwfunc.func_triplet6(xi,xj,xk,xl,xm,xn, gw_prms, det, approximant)
						prmt1 = check_permutations( list(permutations([i,j,k])) ) ; len1 = len(prmt1)
						prmt2 = check_permutations( list(permutations([l,m,n])) ) ; len2 = len(prmt2)
						for n1 in range(len1):
							for n2 in range(len2):
								i, j, k, l, m, n = list(prmt1[n1]) + list(prmt2[n2])
								if(not hide_info):
									print('Triplet(3,3): [%d-%d-%d-%d-%d-%d]' % (i+1,j+1,k+1,l+1,m+1,n+1), end='\r')
								IdxP = i*Np**5 + j*Np**4 + k*Np**3 + l*Np**2 + m*Np + n
								Triplet6[IdxP] += value
			#*****************************#*****************************#*****************************
			#*****************************#*****************************#*****************************

	SNR_sum = np.sqrt(rho2)
	if(not hide_info):
		print("\n >> SNR = ", SNR_sum)
		print(" >> 100/SNR = %.2e" % (100/SNR_sum) + '%\n')

	Doublet = [Doublet3, Doublet4]
	Triplet = [Triplet4, Triplet5, Triplet6]

	return [SNR, GwData, Fisher , Doublet, Triplet]


import numpy as np
import GWDALI.lib.Angles_lib as geo
import GWDALI.lib.Dictionaries as gwdict

rad = np.pi/180
deg = 1./rad

c = 3.e8
R_earth = 6371.e3

Waveforms, PSD, labels_tex = gwdict.Load_Dictionaries()

def Pattern_Func(alpha,beta,psi,shape): # signs fixed
	Coeff = np.sin(shape) #np.sqrt(3.)/2
	fp = Coeff*( 0.5*(1+np.cos(beta)**2)*np.cos(2*alpha) )
	fx = -Coeff*( np.cos(beta)*np.sin(2*alpha) )
	
	Fp =  fp*np.cos(2*psi) + fx*np.sin(2*psi)
	Fx = -fp*np.sin(2*psi) + fx*np.cos(2*psi)
	
	return Fp, Fx

#-------------------------------------------------
def Integral(x,y):
	return np.sum( np.array( [0.5*(y[i]+y[i-1])*(x[i]-x[i-1]) for i in range(1,len(x)) ] ) )

def ScalarProduct(freq,Sn,A,B):
	return 4*np.real( Integral(freq, A*np.conj(B)/Sn) )

def Signal(params, detector, approximant):
	keys = list(params.keys())
	
	#-------------------------------------------------
	# Convert Masses
	#-------------------------------------------------
	if(all( p in  keys for p in ['m1','m2'])):
		m1 = params['m1'] ; m2 = params['m2']
	elif(all( p in keys for p in ['Mc','eta'] )):
		Mc = params['Mc'] ; eta = params['eta'] 
		Coeff = Mc/(2*eta**0.6)
		m1 = Coeff*( 1. + np.sqrt(1.-4*eta) ) # m1>m2
		m2 = Coeff*( 1. - np.sqrt(1.-4*eta) )
	elif(all( p in keys for p in ['Mc','q'] )):
		Mc = params['Mc'] ; q = params['q'] # q = m2/m1
		m2 = ( q*q*(1.+q) )**0.2 * Mc
		m1 = m2/q 
	else: print("\n\n# Invalid options to mass! \n\t >> Available: [m1,m2] or [Mc,eta] or [Mc,q]\n\n")
	#-------------------------------------------------

	alpha    = params['RA']*rad       # rad
	beta     = (90-params['Dec'])*rad # rad
	DL       = params['DL']*1.e3      # Gpc --> Mpc
	iota     = params['iota']         # rad
	psi      = params['psi']          # rad
	t_coal   = params['t_coal']       # sec
	phi_coal = params['phi_coal']     # rad

	sx1 = params['sx1']
	sy1 = params['sy1']
	sz1 = params['sz1']
	sx2 = params['sx2']
	sy2 = params['sy2']
	sz2 = params['sz2']

	s1 = [sx1, sy1, sz1]
	s2 = [sx2, sy2, sz2]

	name  = detector['name']
	freq  = detector['freq'].copy()
	lon   = detector['lon'] # deg
	lat   = detector['lat'] # deg
	rot   = detector['rot']*rad # rad

	phi   = lon*rad
	theta = (90-lat)*rad

	alpha_obs, beta_obs = geo.AngTransf(alpha,beta,theta,phi,rot)
	psi_obs             = geo.poll_ang(alpha,beta,iota,psi,theta,phi,rot)

	t_delay = -np.cos(beta_obs)*R_earth/c

	Fp, Fx    = Pattern_Func(alpha_obs,beta_obs,psi_obs,detector['shape'])
	hp, hx, _ = Waveforms[approximant](m1,m2,iota,DL,s1,s2,freq)

	exp2 = np.exp(-1.j*phi_coal) #
	#exp2 = np.exp( 1.j*( 2*np.pi*freq*(t_coal+t_delay) - phi_coal )  )

	h = Fp*hp + Fx*hx
	
	return h * exp2	

#-------------------------------------------------

eps = 1.e-6
def split_prms(params,x):
	p = params[x]
	dx = np.max([1.e-10,eps*p])
	P1 = params.copy() ; P1[x] = p - dx/2
	P2 = params.copy() ; P2[x] = p + dx/2
	return P1, P2, dx

def Diff1(x,params , detector, approximant):
	P1, P2, dx = split_prms(params,x)
	y2 = Signal(P2, detector, approximant)
	y1 = Signal(P1, detector, approximant)
	return (y2-y1)/dx

def Diff2(xi,xj,params , detector, approximant):
	P1, P2, dx = split_prms(params,xi)
	y2 = Diff1(xj, P2, detector, approximant)
	y1 = Diff1(xj, P1, detector, approximant)
	return (y2-y1)/dx

def Diff3(xi,xj,xk,params , detector, approximant):
	P1, P2, dx = split_prms(params,xi)
	y2 = Diff2(xj,xk, P2, detector, approximant)
	y1 = Diff2(xj,xk, P1, detector, approximant)
	return (y2-y1)/dx

#-------------------------------------------------#-------------------------------------------------
#-------------------------------------------------#-------------------------------------------------

def Fisher_ij(xi,xj, params, detector, approximant): # [1,1]
	diff_xi = Diff1(xi,params , detector, approximant)
	diff_xj = Diff1(xj,params , detector, approximant)
	return ScalarProduct(detector['freq'], detector['Sn'], diff_xi, diff_xj )

#-------------------------------------------------#-------------------------------------------------

def func_doublet3(xi,xj,xk, params, detector, approximant): # [1,2]
	diff_xi    = Diff1(xi,params , detector, approximant)
	diff_xj_xk = Diff2(xj,xk,params , detector, approximant)
	return ScalarProduct(detector['freq'], detector['Sn'], diff_xi, diff_xj_xk)

def func_doublet4(xi,xj,xk,xl, params, detector, approximant): # [2,2]
	diff_xi_xj = Diff2(xi,xj,params , detector, approximant)
	diff_xk_xl = Diff2(xk,xl,params , detector, approximant)
	return ScalarProduct(detector['freq'], detector['Sn'], diff_xi_xj, diff_xk_xl)

#-------------------------------------------------#-------------------------------------------------

def func_triplet4(xi,xj,xk,xl, params, detector, approximant): # [1,3]
	diff_xi       = Diff1(xi,params , detector, approximant)
	diff_xj_xk_xl = Diff3(xj,xk,xl,params , detector, approximant)
	return ScalarProduct(detector['freq'], detector['Sn'], diff_xi, diff_xj_xk_xl)

def func_triplet5(xi,xj,xk,xl,xm, params, detector, approximant): # [2,3]
	diff_xi_xj    = Diff2(xi,xj,params , detector, approximant)
	diff_xk_xl_xm = Diff3(xk,xl,xm,params , detector, approximant)
	return ScalarProduct(detector['freq'], detector['Sn'], diff_xi_xj, diff_xk_xl_xm)

def func_triplet6(xi,xj,xk,xl,xm,xn, params, detector, approximant): # [3,3]
	diff_xi_xj_xk = Diff3(xi,xj,xk,params , detector, approximant)
	diff_xl_xm_xn = Diff3(xl,xm,xn,params , detector, approximant)
	return ScalarProduct(detector['freq'], detector['Sn'], diff_xi_xj_xk, diff_xl_xm_xn)

#-------------------------------------------------#-------------------------------------------------


import numpy as np
from scipy.interpolate import interp1d

try:
	import lal
	import lalsimulation as lalsim
except:
	print('====='*10)
	print('WARNING: lalsuite not instaled!')
	print('Try to install it with: \033[1m conda install -c conda-forge lalsuite \033[0m')
	print('====='*10,'\n')

M_sun = 1.98e30
G = 6.673e-11
c = 3.e8

def integ(x,y): # Integral [Trapezoid Method]
	n = len(x) ; I = 0
	for i in range(1,n): I += 0.5*(y[i]+y[i-1])*(x[i]-x[i-1])
	return I	

#==============================================================================================
# LAL WAVEFORMS
#==============================================================================================

def Waveform_lal(m1, m2, iota, DL, s1, s2, freq, approx):
	dF = np.average(np.diff(freq)) #1 
	f_low, f_max = min(freq), max(freq)

	sx1, sy1, sz1 = s1
	sx2, xy2, sz2 = s2
	
	phi_ref = 0.
	f_ref = 0.

	DL *= 3.086e22
	hp,hc=lalsim.SimInspiralChooseFDWaveform(m1*M_sun,m2*M_sun, sx1, sy1, sz1, sx2, xy2, sz2, DL,
											iota, phi_ref, 0., 0., 0., dF, f_low, f_max, f_ref, None,
											lalsim.GetApproximantFromString(approx))
	
	hp = hp.data.data
	hc = hc.data.data
	
	N = len(hp)
	freqWF = np.linspace(0,dF*(N-1),N)
	
	func_hp = interp1d(freqWF, hp, fill_value=0, kind='linear', bounds_error=False)
	func_hc = interp1d(freqWF, hc, fill_value=0, kind='linear', bounds_error=False)
	
	Hp = func_hp(freq)
	Hx = func_hc(freq)

	hp1 , hp2 = Hp.real, Hp.imag
	hx1 , hx2 = Hx.real, Hp.imag

	hp1 = np.nan_to_num(hp1)
	hx1 = np.nan_to_num(hp2)
	hp2 = np.nan_to_num(hx1)
	hx2 = np.nan_to_num(hx2)

	Hp = hp1 +1.j*hp2
	Hx = hx1 +1.j*hx2
	
	return Hp, Hx, freq

def Waveform_TaylorF2_lal(m1, m2, iota, DL, s1, s2, freq):
	return Waveform_lal(m1, m2, iota, DL, s1, s2, freq, 'TaylorF2')

def Waveform_IMRPhenomD(m1, m2, iota, DL, s1, s2, freq):	
	return Waveform_lal(m1, m2, iota, DL, s1, s2, freq, 'IMRPhenomD')

def Waveform_IMRPhenomP(m1, m2, iota, DL, s1, s2, freq):
	return Waveform_lal(m1, m2, iota, DL, s1, s2, freq, 'IMRPhenomP')

#==============================================================================================
# PYTHON WAVEFORMS ( Leading Order and TaylorF2(3.5PN) )
#==============================================================================================

def GW_Amplitude(m1,m2,DL,freq):
	DL *= 3.086e22     # DL [meters]
	M = m1+m2			 # Total Mass
	eta = m1*m2/M**2   # Symmetric Mass Ratio
	M *= M_sun
	Mc = M*eta**(3./5) # Chirp Mass
	R_isco = 6*(G*M)/c**2
	f_isco = np.sqrt(G*M/R_isco**3)/np.pi # gw frequency at R=R_isco
	Cutoff = np.ones(len(freq))*(freq<4*f_isco)
	return Cutoff*pow(freq,-7./6)*np.sqrt(5./(24*c**3))*pow(np.pi,-2./3)*pow(G*Mc,5./6)/DL
	
def GW_Phase_Simple(m1,m2,freq):
	M = m1+m2			 # Total Mass
	eta = m1*m2/M**2   # Symmetric Mass Ratio
	M *= M_sun
	GM_c3 = G*M/c**3
	t_coal, phi_coal = 0, 0
	phase = 2*np.pi*freq*t_coal - phi_coal + (3./(128*eta))/(np.pi*GM_c3*freq)**(5./3)
	return phase
	
def GW_Phase_TaylorF2(m1,m2,freq):
	M = m1+m2
	eta = m1*m2/M**2
	M *= M_sun
	GM_c3 = G*M/c**3
	
	A = list(np.zeros(8))
	A[0] = 1.
	A[1] = 0.
	A[2] = (3715./756) + (55./9)*eta 
	A[3] = -16*np.pi
	A[4] = (15293365./508032) + (27145*eta/504) + (3085*eta**2/72)
	A[5] = np.pi*((38645./756) - (65*eta/9))*(1.+np.log((6.**1.5)*np.pi*GM_c3*freq))
	A[6] = (11583231236531./4694215680) - (640*np.pi**2/3) - (6848*np.euler_gamma/21) + (-(15737765635./3048192)+(2255*np.pi**2/12))*eta + (76055*eta**2/1728) - (127825*eta**3/1296) - (6848./63)*np.log(64*np.pi*GM_c3*freq)
	A[7] = np.pi*((77096675./254016) + (378515*eta/1512) - (74045*eta**2/756))

	add = np.zeros(len(freq))  # 3.5PN Corrections from idx>0
	for idx in range(0,8):
		add += A[idx]*(np.pi*GM_c3*freq)**((idx-5)/3)
	add *= 3./(128*eta)

	t_coal, phi_coal = 0, 0
	phase = 2*np.pi*freq*t_coal + phi_coal + add

	return phase
	
def Waveform_TaylorF2(m1, m2, iota, DL, s1, s2, freq): # Sathyaprakash-Schutz(2009)
	Amp   = GW_Amplitude(m1,m2,DL,freq)
	phase = GW_Phase_TaylorF2(m1,m2,freq)
	
	exp = np.exp(1.j*(phase-np.pi/4))

	cos_iota = np.cos(iota)
	Gp, Gx = 0.5*(1.+cos_iota*cos_iota), 1.j*cos_iota
	Hplus  = Gp*Amp*exp
	Hcross = Gx*Amp*exp

	return Hplus, Hcross, freq #, Amp, exp
		
def Waveform_Simple(m1, m2, iota, DL, s1, s2, freq):
	Amp = GW_Amplitude(m1,m2,DL,freq)
	phase = GW_Phase_Simple(m1,m2,freq)
	
	cos_iota = np.cos(iota)
	Gp, Gx = 0.5*(1.+cos_iota*cos_iota), cos_iota
	Hplus  = Gp*Amp*np.exp(1.j*(phase-np.pi/4))
	Hcross = Gx*Amp*np.exp(1.j*(phase+np.pi/4))
	return Hplus, Hcross, freq


from __future__ import division
import numpy as np
import bilby, os, shutil, sys
from time import time as now

import GWDALI.lib.Derivatives_Tensors as gwfunc
from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(70,0.3)

#sys.stdout = open(os.devnull, 'w')
bilby.core.utils.setup_logger(log_level=0)	# hide init bilby output in screen
bilby.core.utils.setup_logger(log_level=10)
bilby.core.utils.setup_logger(log_level=20)
bilby.core.utils.setup_logger(log_level=30)
bilby.core.utils.setup_logger(log_level=40)
bilby.core.utils.setup_logger(log_level=50)

#--------------------------------------------------
class GW_likelihood(bilby.Likelihood):
	def __init__(self,args):
		aux = {}
		self.FreeParams = args[0]
		self.Data       = args[1]
		self.GWparams   = args[2]
		self.DetAp		= args[3]
		for f in args[0]: aux[f] = 0
		bilby.Likelihood.__init__(self, aux)
		pass

	def log_likelihood(self):
		Np = len(self.FreeParams)
		params = self.GWparams.copy()
		for fp in self.FreeParams:
			params[fp] = self.parameters[fp]
			if(fp=='Dec'): params[fp] *= 180/np.pi
			
		loglike = 0 ; ndet = 0
		detectors, approximant = self.DetAp 
		for det in detectors:
			h = gwfunc.Signal(params, det, approximant)
			loglike += -gwfunc.ScalarProduct(det['freq'],det['Sn'],self.Data[ndet]-h,self.Data[ndet]-h)
			ndet += 1 
		if(np.isnan(loglike)): loglike = -np.inf
		return loglike
#--------------------------------------------------

class DALI_likelihood(bilby.Likelihood):
	def __init__(self,args):
		self.FreeParams  = args[0]
		self.Theta0      = args[1]
		self.Tensors     = args[2]
		self.dali_method = args[3]

		aux = {}
		for f in args[0]:
			aux[f] = 0

		self.it = 0
		bilby.Likelihood.__init__(self, aux)
		pass

	def log_likelihood(self):
		Theta = [] ; Np = len(self.FreeParams)
		for fp in self.FreeParams:
			prm = self.parameters[fp]
			if(fp=='Dec'): prm *= 180/np.pi
			Theta.append(prm)

		Theta  = np.array(Theta)
		dT = np.array( self.Theta0 - Theta )

		Fisher , Doublet , Triplet = self.Tensors

		Fisher_vec = Fisher.ravel() # convert 2D matrix to 1D array
		Doublet3, Doublet4 = Doublet
		Triplet4, Triplet5, Triplet6 = Triplet

		# dX_ijk:  Tensor rank 3
		# dX_ijkl: Tensor rank 4

		dX2 = np.zeros(Np*Np)
		if(self.dali_method in ['Doublet','Triplet']):
			dX3 = np.zeros(Np**3)
			dX4 = np.zeros(Np**4)
			if(self.dali_method == 'Triplet'):
				dX5 = np.zeros(Np**5)
				dX6 = np.zeros(Np**6)

		#-------------------------------------------------------------
		# For Fisher
		#-------------------------------------------------------------
		for i in range(Np):
			for j in range(Np):
				idx2 = Np*i + j
				dX2[idx2] = dT[i]*dT[j]
				#-------------------------------------------------------------
				# For Doublet
				#-------------------------------------------------------------
				if(self.dali_method in ['Doublet','Triplet']):
					for k in range(Np):
						idx3 = i*Np**2 + j*Np + k
						dX3[idx3] = dT[i]*dT[j]*dT[k]
						for l in range(Np):
							idx4 = i*Np**3 + j*Np**2 + k*Np + l
							dX4[idx4] = dT[i]*dT[j]*dT[k]*dT[l]
							#-------------------------------------------------------------
							# For Triplet
							#-------------------------------------------------------------
							if(self.dali_method == 'Triplet'):
								for m in range(Np):
									idx5 = i*Np**4 + j*Np**3 + k*Np**2 + l*Np + m
									dX5[idx5] = dT[i]*dT[j]*dT[k]*dT[l]*dT[m]
									for n in range(Np):
										idx6 = i*Np**5 + j*Np**4 + k*Np**3 + l*Np**2 + m*Np + n
										dX6[idx6] = dT[i]*dT[j]*dT[k]*dT[l]*dT[m]*dT[n]
		#-------------------------------------------------------------

		loglike = - sum(Fisher_vec*dX2)/2.
		if(self.dali_method in ['Doublet','Triplet']):
			loglike -= (  sum(Doublet3*dX3)/2. + sum(Doublet4*dX4)/8. )
			if(self.dali_method == 'Triplet'):
				loglike -= ( sum(Triplet4*dX4)/6. + sum(Triplet5*dX5)/12. + sum(Triplet6*dX6)/72. )

		if(np.isnan(loglike) or np.isnan(loglike)): loglike = -1.e10
		return loglike

#----------------------------------------
# Prior(DL): dVc/dz
#----------------------------------------
x = np.linspace(1.e-3,10,1000)
d = cosmo.luminosity_distance(x).value
H = cosmo.H(x).value
Rc = d/(1+x)
priorD = 4*np.pi*Rc**2/H
xx = d/1.e3
yy = priorD/sum(priorD)
#----------------------------------------

def get_posterior(FreeParams, Theta0, Detection_Dict, GwData, approximant, Detectors, Tensors, dali_method, sampler_method, npoints):
	Dict = {}
	DL0 = Detection_Dict['DL']
	d1 = cosmo.luminosity_distance(0.001).value / 1.e3 # Gpc
	d2 = cosmo.luminosity_distance(5.0).value / 1.e3   # Gpc
	#----------------------------#----------------------------#----------------------------
	Dict['DL']    = bilby.core.prior.Interped(name='DL',xx=xx,yy=yy,minimum=d1, maximum=d2)
	Dict['iota']  = bilby.core.prior.Sine(name='iota', minimum=0, maximum=np.pi)
	Dict['psi']   = bilby.core.prior.Uniform(name='psi',minimum=0, maximum=np.pi)
	#----------------------------#----------------------------#----------------------------
	Dict['alpha'] = bilby.core.prior.Uniform(name='alpha',minimum=-np.pi, maximum=np.pi)
	Dict['beta']  = bilby.core.prior.Sine(name='beta',minimum=0, maximum=np.pi)
	#----------------------------#----------------------------#----------------------------
	Dict['RA']    = bilby.core.prior.Uniform(name='RA',minimum=-180, maximum=180) # deg unit
	Dict['Dec']   = bilby.core.prior.Cosine(name='Dec',minimum=-np.pi/2, maximum=np.pi/2) # rad unit
	#----------------------------#----------------------------#----------------------------
	Dict['m1']    = bilby.core.prior.Uniform(name='m1',minimum=0.1, maximum=100)
	Dict['m2']    = bilby.core.prior.Uniform(name='m2',minimum=0.1, maximum=100)
	#----------------------------#----------------------------#----------------------------
	Dict['Mc']    = bilby.core.prior.Uniform(name='Mc',minimum=0.1, maximum=100)
	Dict['eta']   = bilby.core.prior.Uniform(name='eta',minimum=1.e-3, maximum=1./4)
	Dict['q']     = bilby.core.prior.Uniform(name='q',minimum=1.e-3, maximum=1.0)
	#----------------------------#----------------------------#----------------------------
	Dict['sx1']    = bilby.core.prior.Uniform(name='sx1',minimum=0, maximum=1.0)
	Dict['sy1']    = bilby.core.prior.Uniform(name='sy1',minimum=0, maximum=1.0)
	Dict['sz1']    = bilby.core.prior.Uniform(name='sz1',minimum=0, maximum=1.0)
	Dict['sx2']    = bilby.core.prior.Uniform(name='sx2',minimum=0, maximum=1.0)
	Dict['sy2']    = bilby.core.prior.Uniform(name='sy2',minimum=0, maximum=1.0)
	Dict['sz2']    = bilby.core.prior.Uniform(name='sz2',minimum=0, maximum=1.0)
	#----------------------------#----------------------------#----------------------------
	Dict['phi_coal']  = bilby.core.prior.Uniform(name='phi_coal',minimum=0, maximum=2*np.pi)
	Dict['t_coal']    = bilby.core.prior.Uniform(name='t_coal',minimum=0, maximum=3600) # 1 hour
	#----------------------------#----------------------------#----------------------------

	Priors = {}
	for fp in FreeParams:
		Priors[fp] = Dict[fp]

	'''
	self.FreeParams = args[0]
	self.Data       = args[1]
	self.GWparams   = args[2]
	self.DetAp		= args[3]
	'''

	if(dali_method=='Standard'):
		likelihood = GW_likelihood([FreeParams, GwData, Detection_Dict, [Detectors,approximant] ])
	else:
		likelihood = DALI_likelihood([FreeParams, Theta0, Tensors, dali_method])

	outdir = 'outdir_%s_%d/' % (dali_method, int(np.random.uniform(0,300))  )
	if(os.path.isdir(outdir)):
		shutil.rmtree(outdir)

	#----------------------------
	# Running Sampler
	#----------------------------
	
	res = bilby.run_sampler(likelihood=likelihood, priors=Priors,
							sampler=sampler_method, npoints=npoints, 
							nsteps=npoints, nwalkers=npoints, 
							nburn=int(0.8*npoints), outdir=outdir)
	shutil.rmtree(outdir)

	#----------------------------

	M = []
	for fp in FreeParams:
		M.append(res.posterior[fp])
	M = np.transpose(np.array(M))

	return M


import numpy as np
from itertools import permutations

def get_indep_indexies(Np,hide_info):
	if(not hide_info):
		print("#-------------------------------------")
		print("Getting Symmetric Indexies ...")

	Doub_12 = np.zeros([Np,Np,Np])
	Doub_22 = np.zeros([Np,Np,Np,Np])
	Trip_13 = np.zeros([Np,Np,Np,Np])
	Trip_23 = np.zeros([Np,Np,Np,Np,Np])
	Trip_33 = np.zeros([Np,Np,Np,Np,Np,Np])

	n3 = Np**3 ; v3 = np.linspace(1,n3,n3)
	n4 = Np**4 ; v4 = np.linspace(1,n4,n4)
	n5 = Np**5 ; v5 = np.linspace(1,n5,n5)
	n6 = Np**6 ; v6 = np.linspace(1,n6,n6)

	cont3, cont4, cont5, cont6 = 0, 0, 0, 0
	for i in range(Np):
		for j in range(Np):
			for k in range(Np):
				#-------------------------------------
				# Doublet
				#-------------------------------------
				Doub_12[i][j][k] = v3[cont3]
				Doub_12[i][k][j] = v3[cont3]

				cont3 += 1
				for l in range(Np):
					prmt1 = list( permutations([i,j]) )
					prmt2 = list( permutations([k,l]) )

					for P1 in range(len(prmt1)):
						for P2 in range(len(prmt2)):
							I, J, K, L = list(prmt1[P1])+list(prmt2[P2])
							Doub_22[I][J][K][L] = v4[cont4]
					
					#-------------------------------------
					# Triplet
					#-------------------------------------

					prmt3 = list( permutations([j,k,l]) )
					for P3 in range(len(prmt3)):
						J, K, L = list(prmt3[P3])
						Trip_13[i][J][K][L] = v4[cont4]

					cont4 += 1
					
					for m in range(Np):
						prmt1 = list( permutations([i,j]) )
						prmt2 = list( permutations([k,l,m]) )

						for P1 in range(len(prmt1)):
							for P2 in range(len(prmt2)):
								I, J, K, L, M = list(prmt1[P1])+list(prmt2[P2])
								Trip_23[I][J][K][L][M] = v5[cont5]

						cont5 += 1

						for n in range(Np):
							prmt1 = list( permutations([i,j,k]) )
							prmt2 = list( permutations([l,m,n]) )

							for P1 in range(len(prmt1)):
								for P2 in range(len(prmt2)):
									I, J, K, L, M, N = list(prmt1[P1])+list(prmt2[P2])
									Trip_33[I][J][K][L][M][N] = v6[cont6]

							cont6 += 1

	vals1, vals2, vals3, vals4, vals5 = [], [], [], [], []
	indepD_12 = []
	indepD_22 = []
	indepT_13 = []
	indepT_23 = []
	indepT_33 = []

	for i in range(Np):
		for j in range(Np):
			for k in range(Np):
				v1 = Doub_12[i][j][k]
				if(not v1 in vals1):
					vals1.append(v1)
					indepD_12.append([i,j,k])

				for l in range(Np):
					v2 = Doub_22[i][j][k][l]
					if(not v2 in vals2):
						vals2.append(v2)
						indepD_22.append([i,j,k,l])

					v3 = Trip_13[i][j][k][l]
					if(not v3 in vals3):
						vals3.append(v3)
						indepT_13.append([i,j,k,l])

					for m in range(Np):
						v4 = Trip_23[i][j][k][l][m]
						if(not v4 in vals4):
							vals4.append(v4)
							indepT_23.append([i,j,k,l,m])

						for n in range(Np):
							v5 = Trip_33[i][j][k][l][m][n]
							if(not v5 in vals5):
								vals5.append(v5)
								indepT_33.append([i,j,k,l,m,n])

	idxs_doub = [indepD_12, indepD_22]
	idxs_trip = [indepT_13, indepT_23, indepT_33]

	if(not hide_info):
		print("# [Ntot - Nindep]:\n")
		print('Doub(1,2): ', n3, len(indepD_12))
		print('Doub(2,2): ', n4, len(indepD_22))
		print('Trip(1,3): ', n4, len(indepT_13))
		print('Trip(2,3): ', n5, len(indepT_23))
		print('Trip(2,3): ', n6, len(indepT_33))
		print("#----------"*3)

	return idxs_doub, idxs_trip



import numpy as np
import matplotlib.pyplot as plt

import os, shutil, sys
import warnings ; warnings.filterwarnings("ignore")

#===========================
#----Internal_libs----------
#===========================
import GWDALI.lib.Waveforms as wf
import GWDALI.lib.Angles_lib as geo
import GWDALI.lib.Dictionaries as gwdict
import GWDALI.lib.Derivatives_Tensors as gwfunc
import GWDALI.lib.Corner_Plots as corner

from .lib.Likelihood import get_posterior
from .lib.Compute_Tensors import Get_Tensors

#===========================

from scipy.interpolate import interp1d
from astropy.cosmology import FlatLambdaCDM
from time import time as now
from tqdm import trange

cosmo = FlatLambdaCDM(70,0.3)

#-------------------------------------------------
# From Dictionaries.py
Waveforms, PSD, labels_tex = gwdict.Load_Dictionaries()
#-------------------------------------------------
freq0 = 10**np.linspace(0,4,4000)

'''
#-------------------------------------------------------
# CAUTION! From GWFISH 
#-------------------------------------------------------
def invertSVD(matrix): 
    thresh = 1e-10

    dm = np.sqrt(np.diag(matrix))
    normalizer = np.outer(dm, dm)
    matrix_norm = matrix / normalizer

    [U, S, Vh] = np.linalg.svd(matrix_norm)

    kVal = sum(S > thresh)
    matrix_inverse_norm = U[:, 0:kVal] @ np.diag(1. / S[0:kVal]) @ Vh[0:kVal, :]

    # print(matrix @ (matrix_inverse_norm / normalizer))

    return matrix_inverse_norm / normalizer
#-------------------------------------------------------
'''

#-------------------------------------------------------
# Numpy routine: 
#	"Compute the (Moore-Penrose) pseudo-inverse of a matrix"
#-------------------------------------------------------
def invert_matrix(matrix, rcond): 

    dm = np.sqrt(np.diag(matrix))
    norm = np.outer(dm, dm)
    M_norm = matrix / norm

    M_inv = np.linalg.pinv(M_norm,rcond=rcond)

    return M_inv / norm
#-------------------------------------------------------

def GWDALI( Detection_Dict, 
			FreeParams, 
			detectors, 
			approximant = 'TaylorF2',
			fmin  = 1., 
			fmax  = 1.e4, 
			fsize = 3000, 
			dali_method    = 'Doublet',
			sampler_method = 'nestle', 
			npoints      = 100,
			rcond		 = 1.e-4,
			save_samples = False, 
			save_cov     = False, 
			save_fisher  = False,
			plot_corner  = False,
			hide_info    = False,
			index		 = 1):

	if(not hide_info):
		print('\n'+"-----"*10)
		print("-------------- Running GWDALI --------------------")
		print("-----"*10+'\n')
		print(" >> FreeParams: " ,  FreeParams)
		print(" >> dali_method: %s" % dali_method)
		print(" >> approximant: %s" % approximant)
		print(" >> sampler_method: %s\n" % sampler_method)

	T1 = now()
	
	cond_out = any([save_cov, save_samples, save_fisher, plot_corner])
	#-----------------------------------------------------------------------------------
	# Cria uma pasta para armazenar os resultados
	if(cond_out):
		path = 'output_%s/' % dali_method
		if(not os.path.isdir(path)): os.mkdir(path)
	#-----------------------------------------------------------------------------------

	Np   = len(FreeParams)
	keys = Detection_Dict.keys()
	freq = 10**np.linspace(np.log10(fmin), np.log10(fmax), fsize)

	#-------------------------------------------------
	# Atribuir ao detector seu respectivo Sn(f)
	for idx in range(len(detectors)):
		det = detectors[idx]
		name = det['name']
		Sn0 = PSD[name]
		func_Sn = interp1d(freq0, Sn0, fill_value=np.inf, bounds_error=False)
		detectors[idx]['freq'] = freq
		detectors[idx]['Sn']   = func_Sn(freq)

	t_init = now()
	#-----------------------------------------------------------------------------------
	
	Result = {}

	#=====================================================
	# Armazena os valores das injeções (Dict)
	truths = {}
	for key in Detection_Dict.keys():
		truths[key] = Detection_Dict[key]
	#=====================================================
	# Armazena os valores das injeções (List)
	Theta0 = []
	for fp in FreeParams:
		Theta0.append(Detection_Dict[fp])
	Theta0 = np.array(Theta0)
	#=====================================================
	
	t1 = now()
	#-----------------------------------------------------------------------------------
	# From Compute_Tensors.py (looping in detectors)
	#-----------------------------------------------------------------------------------
	SNR , GwData, Fisher , Doublet , Triplet = Get_Tensors(FreeParams, Detection_Dict, detectors, dali_method, approximant,hide_info)
	Tensors = [Fisher , Doublet , Triplet]
	#-----------------------------------------------------------------------------------
	t2 = now()
	dt = t2 - t1
	if(not hide_info):
		print('\n'+"-----"*10)
		print("\n>> Fisher/Doublet/Triplet computed in %d sec" %dt)
		print('\n'+"-----"*10)

	#********************************************************************

	#------------------------------------------------------
	# Invert Fisher Matrix
	#------------------------------------------------------
	
	Cov = invert_matrix(Fisher, rcond) # from numpy.linalg.pinv

	Result['Fisher']    = Fisher
	Result['CovFisher'] = Cov

	if(not hide_info):
		print("\n Fisher = \n", np.matrix(Fisher),'\n')
		print("\n Covariance = \n", np.matrix(Cov),'\n')
		print("-----"*10+'\n')

	#------------------------------------------------------
	#------------------------------------------------------
	# Salvar as injeções nas headers de cada file:
	#------------------------------------------------------
	header01 = 'FullInjections:\n'
	header02 = ''
	for key in keys:
		header01 += "%s\t" % key
		header02 += "%e\t" % Detection_Dict[key]
	header1 = 'Injections: ' ; header2 = '' ; line = '---'*10
	for fp in FreeParams:
		header1 += '%s\t' % fp
		header2 += '%e\t' % Detection_Dict[fp]
	header = header01 + '\n' + header02 + '\n' + line + '\n' + header1+'\n'+header2 +'\n'+ line

	if(save_fisher): np.savetxt(path+'Fisher_Matrix_%d.txt'%(index),Fisher,fmt='%e',delimiter='\t',header=header)
	if(save_cov): np.savetxt(path+'FisherCov_%d.txt'%(index),Cov,fmt='%e',delimiter='\t',header=header)

	#------------------------------------------------------#------------------------------------------------------
	#-------------------------------------------------MONTE-CARLO-------------------------------------------------
	#------------------------------------------------------#------------------------------------------------------
	vec = [ truths[fp] for fp in FreeParams]
	truths = vec		
	#------------------------------------------------------
	# From Likelihood.py
	#------------------------------------------------------
	if(dali_method != 'Fisher'):
		# allowed only for Fisher_Sampling and Doublet
		if(not hide_info): print("Running Sampler ...\n")
		#------------------------------------------------------#------------------------------------------------------
		M = get_posterior(FreeParams, Theta0, Detection_Dict, GwData, approximant, detectors, Tensors, dali_method, sampler_method, npoints)
		#------------------------------------------------------#------------------------------------------------------
		Cov2 = np.matrix( np.cov(np.transpose(M)) )
		header3 = header + '\n' + line + '\nRecovery:\n'
		for i in range(len(FreeParams)): header3 += '%e\t' % np.average(M[:,i])
		header3 += '\n' + line
		if(save_cov): np.savetxt(path+'Covariance_Matrix_%d.txt'%(index),Cov2,fmt='%e',delimiter='\t',header=header3)

		qs = [0.2,0.5,0.8] # CL = 60%
		Error, Recovery = [], []
		for m in M.T:
			q = np.quantile(m,qs)
			Recovery.append(q[1])
			Error.append(0.5*(q[2]-q[0]))

		Result['Covariance'] = Cov2
		Result['Samples']    = M
		Result['Error']      = Error
		Result['Recovery']   = Recovery
	Result['SNR']		 = SNR
	#------------------------------------------------------
	# Save Samples
	#------------------------------------------------------

	T2 = now()
	dT = T2 - T1
	dT_min = dT // 60
	dT_sec = dT % 60
	title = "Concluded in [%d min : %d sec]" % (dT_min, dT_sec)

	if(save_samples and (dali_method!='Fisher')):
		if(not hide_info): print(">> Saving Samples ...")
		np.savetxt(path+'samples_%d.txt' % (index),M,fmt='%e',delimiter='\t',header=header)

	#------------------------------------------------------
	# Plot Corner
	#------------------------------------------------------

	if(plot_corner):
		labels = [ labels_tex[fp] for fp in FreeParams]

		if(dali_method == 'Fisher'):
			fig = corner.Plot_Fisher(truths, labels, FreeParams, Cov)
			plt.suptitle(title)
			fig.subplots_adjust(wspace=0.01, hspace=0.01)
			fig.savefig(path+'corner_%d.png' % (index))
		else:
			print(">>> Plotting Corner [Doublet] .... ")
			fig = corner.Plot_corner(M, truths, labels, FreeParams, Cov)
			plt.suptitle(title)
			fig.legend(loc='upper right', bbox_to_anchor=(0.95, 0.95))
			fig.subplots_adjust(wspace=0.01, hspace=0.01)
			fig.savefig(path+'corner_%d.png' % (index))	
		plt.show()

	# Result (dict)
	# keys: Fisher, CovFisher, Covariance, Samples, Error, Recovery, SNR
	return Result 